package tacv.universe.map;

import java.util.*;
import javax.vecmath.Point3d;
import javax.vecmath.Tuple3d;
import org.apache.log4j.Logger;
import org.jgrapht.Graphs;
import org.jgrapht.alg.NeighborIndex;
import org.jgrapht.graph.DefaultWeightedEdge;
import org.jgrapht.graph.ListenableUndirectedWeightedGraph;
import org.jgrapht.util.ArrayUnenforcedSet;
import tacv.universe.environment.AgentScoutEnvironment;
import tacv.universe.map.xml.Edge;
import tacv.universe.map.xml.Map;
import tacv.universe.map.xml.Node;
import tacv.util.Point;
import tacv.util.Vector;

/**
 * A graph representing the streets and the cross-roads in the AgentScout village.
 */
public class StreetGraph extends ListenableUndirectedWeightedGraph<MapPosition, DefaultWeightedEdge> {
    private static final Logger LOGGER = Logger.getLogger(StreetGraph.class);
    private static final long serialVersionUID = 7530622762977059100L;

    private static final double NORMALIZED_EDGE_LENGTH = 25.0;

    private AgentScoutEnvironment environment;
    private NeighborIndex<MapPosition, DefaultWeightedEdge> neighborIndex;
    private HashMap<MapPosition, LinkedList<MapPosition>> crossroadNeighbors;

//        /** A* distance between two points */
//    private int[][] d;
//
//    /** Distance from point A to the closest point from which point B is visible */
//    private int[][] v;

    public StreetGraph(AgentScoutEnvironment environment) {
        super(DefaultWeightedEdge.class);

        this.environment = environment;

        // neighbor index
        neighborIndex = new NeighborIndex<MapPosition, DefaultWeightedEdge>(this);
        addGraphListener(neighborIndex);
//        visibilityNeighbors = new HashMap<MapPosition, List<MapPosition>>();
    }

    public NeighborIndex<MapPosition, DefaultWeightedEdge> getNeighborIndex() {
        return neighborIndex;
    }

    public MapPosition getRandomNode() {
        List<MapPosition> vertices = new ArrayList<MapPosition>(vertexSet());
        return vertices.get(environment.getRandom().nextInt(vertices.size()));
    }


    /**
     * Adds a new edge in both directions to crossroadNeighbors.
     * @param from
     * @param to
     */
    private void addCrossroadEdge(MapPosition from, MapPosition to){
        if (crossroadNeighbors == null){
            crossroadNeighbors = new HashMap<MapPosition, LinkedList<MapPosition>>(vertexSet().size());
        }
        LinkedList<MapPosition> neighFrom = crossroadNeighbors.get(from);
        if (neighFrom == null){
            neighFrom = new LinkedList<MapPosition>();
            crossroadNeighbors.put(from, neighFrom);
        }
        if (!neighFrom.contains(to)) neighFrom.add(to);
    }

    /**
     * Adds interpolated nodes the graph in order to make the edges of roughly the same length.
     *
     * @param desiredLength the desired length of the edge (step)
     **/
    public void normalizeEdgeLength(double desiredLength){
        Set<DefaultWeightedEdge> edges = new ArrayUnenforcedSet<DefaultWeightedEdge>(this.edgeSet());

        for(DefaultWeightedEdge edge : edges) {

            MapPosition source = this.getEdgeSource(edge);
            MapPosition target = this.getEdgeTarget(edge);

            addCrossroadEdge(source, target);
            addCrossroadEdge(target, source);

            this.removeEdge(edge);

            double length = source.distance(target);
            double n = Math.round(length / desiredLength);
            double actualLength = length/n;

            Vector vector = new Vector(target);
            vector.sub(source);
            vector.normalize();

            MapPosition lastNode = source;

            // Generate interpolated nodes
            for (int i=1; i < n; i++) {
                Tuple3d p = new Point3d(vector);
                p.scaleAdd(i*actualLength,source.getPosition());
                MapPosition newNode = new MapPosition(source.getId()+"_"+target.getId()+"_"+i, new Point(p));
                this.addVertex(newNode);
                DefaultWeightedEdge newEdge = this.addEdge(lastNode, newNode);
                this.setEdgeWeight(newEdge, actualLength);
                addCrossroadEdge(newNode, source);
                addCrossroadEdge(newNode, target);
                lastNode = newNode;
            }

            // Adds the last edge (if the the loop above does nothing, it recreates the original edge)
            DefaultWeightedEdge lastEdge = this.addEdge(lastNode, target);
            this.setEdgeWeight(
                    lastEdge,
                    lastNode.getPosition().distance(target)
                );


        }
    }

    public MapPosition getMapPosition(String id){
        for (MapPosition mp : vertexSet())
            if (mp.getId().equals(id)) return mp;
        return null;
    }

   
//    private void computeDifNotCached(String basename) {
//        File df = new File(basename + ".d");
//        if (df.exists()) {
//            d = (int[][]) ObjectSerializer.load(df.getAbsolutePath());
//            if (d.length == vertexSet().size()){
//                LOGGER.info("Using cached D values.");
//                return;
//            }
//        }
//        LOGGER.info("Building new D cahe.");
//        computeD();
//        ObjectSerializer.save(df.getAbsolutePath(), d);
//    }

//    //Smart targets interface implementation.
//    public List<IMapNode> getMobilityNeighbors(IMapNode node, IAgentInfo agentType) {
//        return (List) getNeighborIndex().neighborListOf((MapPosition)node);
//    }
//
//    //Smart targets interface implementation.
//    public List<IMapNode> getCrossroadNeighbors(IMapNode node, IAgentInfo agentType){
//        return (List) crossroadNeighbors.get((MapPosition)node);
//    }
//
//    public List<MapPosition> getVisibilityNeighborsUnlimited(MapPosition node) {
//        return visibilityNeighbors.get((MapPosition)node);
//    }

    /**
     * Returns precomputed d-value (a* distance) between two nodes.
     */
//    public int getCachedD(IMapNode p1, IMapNode p2) {
//        return d[((MapPosition)p1).getIntId()][((MapPosition)p2).getIntId()];
//    }

    public void loadFromMarshalledMap(Map map) {
        // A map from node ids to nodes
        ArrayList<MapPosition> nodes = new ArrayList<MapPosition>();

        for (Node n : map.getStreetGraph().getNodes().getNode()) {
            MapPosition pos = new MapPosition(Integer.toString(n.getId()),new Point(n.getX(), n.getY(), n.getZ()));
            addVertex(pos);
            nodes.add(n.getId(), pos);
        }

        for (Edge xmlEdge : map.getStreetGraph().getEdges().getEdge()) {
            MapPosition p1 = nodes.get(xmlEdge.getFrom());
            MapPosition p2 = nodes.get(xmlEdge.getTo());
            addEdge(p1, p2);
        }


        normalizeEdgeLength(NORMALIZED_EDGE_LENGTH);

        System.out.println("Loaded: " + nodes.size() + " nodes");
        System.out.println("Loaded: " + edgeSet().size() + " edges");
//        computeDifNotCached("STDataCache");
    }

//    /**  Precomputes the values of <tt>d</tt>.
//     * */
//    private void computeD(){
//        //compute d
//        d = new int[vertexSet().size()][vertexSet().size()];
//        for (MapPosition p1 : vertexSet()){
//            BellmanFordShortestPath p1dist = new BellmanFordShortestPath(this, p1);
//            for (MapPosition p2 : vertexSet()){
//                if (p1 == p2){
//                    d[p1.getIntId()][p2.getIntId()] = 0;
//                } else {
//                    d[p1.getIntId()][p2.getIntId()] = (int) p1dist.getCost(p2);
//                }
//            }
//        }
//    }

    /** Returns the closest node to the given point. */
    public MapPosition getNearestNode(Point point) {
        MapPosition closest = null;
        for (MapPosition node : vertexSet()){
            if (closest == null || node.distance(point) < closest.distance(point)) {
                closest = node;
            }
        }
        return closest;
    }

//    public MapPosition getNodeByTerm(String name) {
//        for (MapPosition n : vertexSet()) {
//            if (name.equals(n.getTerm()))
//                return n;
//        }
//        return null;
//    }

    public MapPosition getRandomNeighborOfNode(MapPosition node) {
        List<MapPosition> neighbors = Graphs.neighborListOf(this, node);
        return neighbors.get(environment.getRandom().nextInt(neighbors.size()));
    }

    public MapPosition getFarthestNeighborFrom(Point point, MapPosition node) {
        List<MapPosition> neighbors = Graphs.neighborListOf(this, node);
        MapPosition farthestNode = null;
        for (MapPosition neighbor : neighbors) {
            if (farthestNode == null || neighbor.distance(point) > farthestNode.distance(point))
            farthestNode = neighbor;
        }
        return farthestNode;
    }

//    public void computeVisibilityNeighbors(Visibility visibilityModel){
//            double visRadius = TacticalPerson.VISIBILITY_RANGE;
//            for (MapPosition p1 : vertexSet()){
//                ArrayList<MapPosition> visNeighbors = new ArrayList<MapPosition>();
//                visibilityNeighbors.put(p1,visNeighbors);
//                for (MapPosition p2 : vertexSet()){
//                    if (visibilityModel.isVisible(p1, p2, visRadius)) visNeighbors.add(p2);
//               }
//            }
//    }
//
//    public int[] getCachedDistances(IMapNode p1) {
//        return d[((MapPosition)p1).getIntId()];
//    }
}
