package tacv.vis3d.generator;

import com.jme3.material.Material;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Mesh;
import com.jme3.scene.Node;
import com.jme3.scene.VertexBuffer.Type;
import com.jme3.util.BufferUtils;
import java.util.ArrayList;
import java.util.List;
import javax.vecmath.Point2d;
import tacv.universe.world.Terrain;

public class PathVisualization {

    private final static float PATH_HEIGHT_OFFSET = 0.3f;
    private Terrain terrain;

    public PathVisualization(Terrain terrain) {
        this.terrain = terrain;
    }

    public Node buildPathMesh(List<Point2d> path, float pathWidth, float maxSegmentLength, Material material) {
        if (path.size() < 2) {
            throw new IllegalArgumentException("The path.size() has to be >= 2!");
        }

        List<List<Point2d>> subpaths = splitToSubpaths(path);
        Node node = new Node();
        for (List<Point2d> subpath : subpaths) {
            Mesh mesh = buildSubpathMesh(subpath, pathWidth, maxSegmentLength);
            Geometry geom = new Geometry("pathMesh", mesh);
            geom.setMaterial(material);
            node.attachChild(geom);
        }
        return node;
    }

    private List<List<Point2d>> splitToSubpaths(List<Point2d> path) {
        List<List<Point2d>> split = new ArrayList<List<Point2d>>();
        List<Point2d> part = new ArrayList<Point2d>();
        split.add(part);
        part.add(path.get(0));
        part.add(path.get(1));
        for (int i = 2; i < path.size(); i++) {
            Point2d p1 = part.get(part.size() - 2);
            Point2d p2 = part.get(part.size() - 1);
            Point2d p3 = path.get(i);
            double ux = p2.x - p1.x;
            double uy = p2.y - p1.y;
            double vx = p3.x - p2.x;
            double vy = p3.y - p2.y;
            if (ux * vx + uy * vy >= 0) {
                part.add(p3);
            } else {
                part = new ArrayList<Point2d>();
                split.add(part);
                part.add(p2);
                part.add(p3);
            }
        }
        return split;
    }

    private Mesh buildSubpathMesh(List<Point2d> subpath, float pathWidth, float maxSegmentLength) {
        int segmentCount = subpath.size() - 1;

        Vector3f[] vertices = new Vector3f[(segmentCount + 1) * 2];
        int vertexID = 0;
        Vector3f[] prevPoly = null;
        for (int segmentID = 0; segmentID < segmentCount; segmentID++) {
            Vector3f[] currPoly = calculatePathPolygon(subpath.get(segmentID), subpath.get(segmentID + 1), pathWidth);
            if (prevPoly == null) {
                vertices[vertexID + 0] = currPoly[0];
                vertices[vertexID + 1] = currPoly[1];
            } else {
                Vector3f q1 = calculateIntersection(prevPoly[0], prevPoly[2], currPoly[0], currPoly[2]);
                Vector3f q2 = calculateIntersection(prevPoly[1], prevPoly[3], currPoly[1], currPoly[3]);
                vertices[vertexID + 0] = q1 != null ? q1 : currPoly[0];
                vertices[vertexID + 1] = q2 != null ? q2 : currPoly[1];
            }
            vertexID += 2;
            if (segmentID == segmentCount - 1) {
                vertices[vertexID + 0] = currPoly[2];
                vertices[vertexID + 1] = currPoly[3];
            }
            prevPoly = currPoly;
        }

        vertices = subdividePath(vertices, maxSegmentLength);
        segmentCount = vertices.length / 2 - 1;

        for (Vector3f vertex : vertices) {
            vertex.z = PATH_HEIGHT_OFFSET + terrain.getAltitude(vertex.x, vertex.y);
        }

        int[] indices = new int[segmentCount * 2 * 3];
        for (int segmentID = 0; segmentID < segmentCount; segmentID++) {
            int j = segmentID * 2;
            int i = j * 3;
            indices[i + 0] = j + 0;
            indices[i + 1] = j + 1;
            indices[i + 2] = j + 3;
            indices[i + 3] = j + 0;
            indices[i + 4] = j + 3;
            indices[i + 5] = j + 2;
        }

        Mesh mesh = new Mesh();
        mesh.setBuffer(Type.Position, 3, BufferUtils.createFloatBuffer(vertices));
        //mesh.setBuffer(Type.TexCoord, 2, BufferUtils.createFloatBuffer(texCoord));
        mesh.setBuffer(Type.Index, 1, BufferUtils.createIntBuffer(indices));

        return mesh;
    }

    private Vector3f[] subdividePath(Vector3f[] vertices, double maxSegmentLength) {
        List<Vector3f> vertexList = new ArrayList<Vector3f>();
        int segmentCount = vertices.length / 2 - 1;
        for (int segmentID = 0; segmentID < segmentCount; segmentID++) {
            Vector3f p0 = vertices[segmentID * 2];
            Vector3f p1 = vertices[segmentID * 2 + 1];
            Vector3f p2 = vertices[(segmentID + 1) * 2];
            Vector3f p3 = vertices[(segmentID + 1) * 2 + 1];
            double ux = p2.x - p0.x;
            double uy = p2.y - p0.y;
            double vx = p3.x - p1.x;
            double vy = p3.y - p1.y;
            double ulen = Math.sqrt(ux * ux + uy * uy);
            double vlen = Math.sqrt(vx * vx + vy * vy);
            int newSegCount = (int) Math.max(
                    Math.ceil(ulen / maxSegmentLength),
                    Math.ceil(vlen / maxSegmentLength));
            vertexList.add(p0);
            vertexList.add(p1);
            double dt = 1.0 / newSegCount;
            double t = 0;
            for (int i = 0; i < newSegCount - 1; i++) {
                t += dt;
                vertexList.add(new Vector3f((float) (p0.x + ux * t), (float) (p0.y + uy * t), 0));
                vertexList.add(new Vector3f((float) (p1.x + vx * t), (float) (p1.y + vy * t), 0));
            }
            if (segmentID == segmentCount - 1) {
                vertexList.add(p2);
                vertexList.add(p3);
            }
        }
        Vector3f[] vertexArray = new Vector3f[vertexList.size()];
        vertexList.toArray(vertexArray);
        return vertexArray;
    }

    private Vector3f[] calculatePathPolygon(Point2d startPoint, Point2d endPoint, float pathWidth) {
        double ux = endPoint.x - startPoint.x;
        double uy = endPoint.y - startPoint.y;
        double ulen = Math.sqrt(ux * ux + uy * uy);
        double k = pathWidth * 0.5 / ulen;
        ux *= k;
        uy *= k;
        Vector3f p1 = new Vector3f((float) (startPoint.x - uy), (float) (startPoint.y + ux), 0);
        Vector3f p2 = new Vector3f((float) (startPoint.x + uy), (float) (startPoint.y - ux), 0);
        Vector3f p3 = new Vector3f((float) (endPoint.x - uy), (float) (endPoint.y + ux), 0);
        Vector3f p4 = new Vector3f((float) (endPoint.x + uy), (float) (endPoint.y - ux), 0);
        return new Vector3f[]{p1, p2, p3, p4};
    }

    private Vector3f calculateIntersection(Vector3f line1Start, Vector3f line1End, Vector3f line2Start, Vector3f line2End) {
        float x13 = line1Start.x - line2Start.x;
        float x21 = line1End.x - line1Start.x;
        float x43 = line2End.x - line2Start.x;
        float y13 = line1Start.y - line2Start.y;
        float y21 = line1End.y - line1Start.y;
        float y43 = line2End.y - line2Start.y;
        float z13 = line1Start.z - line2Start.z;
        float z21 = line1End.z - line1Start.z;
        float z43 = line2End.z - line2Start.z;
        float d4343 = x43 * x43 + y43 * y43 + z43 * z43;
        if (d4343 == 0) {
            return null;
        }
        float d2121 = x21 * x21 + y21 * y21 + z21 * z21;
        float d4321 = x43 * x21 + y43 * y21 + z43 * z21;
        float d = d2121 * d4343 - d4321 * d4321;
        if (d == 0) {
            return null;
        }
        float d1343 = x13 * x43 + y13 * y43 + z13 * z43;
        float d1321 = x13 * x21 + y13 * y21 + z13 * z21;
        float t1 = (d1343 * d4321 - d1321 * d4343) / d;
        float t2 = (d1343 + t1 * d4321) / d4343;
        float x1 = line1Start.x + t1 * x21;
        float y1 = line1Start.y + t1 * y21;
        float z1 = line1Start.z + t1 * z21;
        float x2 = line2Start.x + t2 * x43;
        float y2 = line2Start.y + t2 * y43;
        float z2 = line2Start.z + t2 * z43;
        float x = (x1 + x2) * 0.5f;
        float y = (y1 + y2) * 0.5f;
        float z = (z1 + z2) * 0.5f;
        return new Vector3f(x, y, z);
    }
}
